(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
	typeof define === 'function' && define.amd ? define(factory) :
	(global = global || self, global.pouchVue = factory());
}(this, function () { 'use strict';

	function createCommonjsModule(fn, module) {
		return module = { exports: {} }, fn(module, module.exports), module.exports;
	}

	var src = createCommonjsModule(function (module) {
	(function() {
	    var vue = null,
	        pouch = null,
	        defaultDB = null,
	        defaultUsername = null,
	        defaultPassword = null,
	        databases = {},
	        optionsDB = {};

	    var vuePouch = {
	        // lifecycle hooks for mixin
	        beforeDestroy: function beforeDestroy() {
	            Object.values(this._liveFeeds).map(function (lf) {
	                lf.cancel();
	            });
	        },
	        created: function created() {
	            if (!vue) {
	                console.warn('pouch-vue not installed!');
	                return;
	            }

	            var defineReactive = vue.util.defineReactive,
	                vm = this;

	            vm._liveFeeds = {};

	            if (defaultDB) {
	                databases[defaultDB] = new pouch(defaultDB, optionsDB);
	                registerListeners(databases[defaultDB]);
	            }

	            function fetchSession(db) {
	                if ( db === void 0 ) db = databases[defaultDB];

	                return new Promise(function (resolve) {
	                    db
	                        .getSession()
	                        .then(function (session) {
	                            db
	                                .getUser(session.userCtx.name)
	                                .then(function (userData) {
	                                    var userObj = Object.assign(
	                                        {},
	                                        session.userCtx,
	                                        userData
	                                    );
	                                    resolve({
	                                        user: userObj,
	                                        hasAccess: true,
	                                    });
	                                })
	                                .catch(function (error) {
	                                    resolve(error);
	                                });
	                        })
	                        .catch(function (error) {
	                            resolve(error);
	                        });
	                });
	            }

	            function login(db) {
	                if ( db === void 0 ) db = databases[defaultDB];

	                return new Promise(function (resolve) {

	                    db
	                        .login(defaultUsername, defaultPassword)
	                        .then(function (user) {
	                            db
	                                .getUser(user.name)
	                                .then(function (userData) {
	                                    var userObj = Object.assign(
	                                        {},
	                                        user,
	                                        userData
	                                    );
	                                    resolve({
	                                        user: userObj,
	                                        hasAccess: true,
	                                    });
	                                })
	                                .catch(function (error) {
	                                    resolve(error);
	                                });
	                        })
	                        .catch(function (error) {
	                            resolve(error);
	                        });
	                });
	            }

	            function makeInstance(db, opts) {
	                if ( opts === void 0 ) opts = {};

	                databases[db] = new pouch(db, opts);
	                registerListeners(databases[db]);
	            }

	            function registerListeners(db) {
	                db.on('created', function (name) {
	                    vm.$emit('pouchdb-db-created', {
	                        db: name,
	                        ok: true,
	                    });
	                });
	                db.on('destroyed', function (name) {
	                    vm.$emit('pouchdb-db-destroyed', {
	                        db: name,
	                        ok: true,
	                    });
	                });
	            }

	            var $pouch = {
	                version: '0.1.37',
	                connect: function connect(username, password, db) {
	                    if ( db === void 0 ) db = databases[defaultDB];

	                    return new Promise(function (resolve) {
	                        defaultUsername = username;
	                        defaultPassword = password;

	                        if (!db._remote) {
	                            resolve({
	                                message: 'database is not remote',
	                                error: 'bad request',
	                                status: 400,
	                            });
	                            return;
	                        }

	                        login(db).then(function (res) {
	                            resolve(res);
	                        });
	                    });
	                },
	                createUser: function createUser(username, password, db) {
	                    if ( db === void 0 ) db = databases[defaultDB];

	                    return db
	                        .signup(username, password)
	                        .then(function () {
	                            return vm.$pouch.connect(username, password, db);
	                        })
	                        .catch(function (error) {
	                            return new Promise(function (resolve) {
	                                resolve(error);
	                            });
	                        });
	                },
	                putUser: function putUser (username, metadata, db) {
	                    if ( metadata === void 0 ) metadata = {};
	                    if ( db === void 0 ) db = databases[ defaultDB ];

	                    return db
	                        .putUser(username, {
	                            metadata: metadata
	                        })
	                        .catch(function (error) {
	                            return new Promise(function (resolve) {
	                                resolve(error);
	                            });
	                        });
	                },
	                deleteUser: function deleteUser (username, db) {
	                    if ( db === void 0 ) db = databases[ defaultDB ];

	                    return db
	                        .deleteUser(username)
	                        .catch(function (error) {
	                            return new Promise(function (resolve) {
	                                resolve(error);
	                            });
	                        });
	                },
	                changePassword: function changePassword (username, password, db) {
	                    if ( db === void 0 ) db = databases[ defaultDB ];

	                    return db
	                        .changePassword(username, password)
	                        .catch(function (error) {
	                            return new Promise(function (resolve) {
	                                resolve(error);
	                            });
	                        });
	                },
	                changeUsername: function changeUsername (oldUsername, newUsername, db) {
	                    if ( db === void 0 ) db = databases[ defaultDB ];

	                    return db
	                        .changeUsername(oldUsername, newUsername)
	                        .catch(function (error) {
	                            return new Promise(function (resolve) {
	                                resolve(error);
	                            });
	                        });
	                },
	                signUpAdmin: function signUpAdmin (adminUsername, adminPassword, db) {
	                    if ( db === void 0 ) db = databases[ defaultDB ];

	                    return db
	                        .signUpAdmin(adminUsername, adminPassword)
	                        .catch(function (error) {
	                            return new Promise(function (resolve) {
	                                resolve(error);
	                            });
	                        });
	                },
	                deleteAdmin: function deleteAdmin (adminUsername, db) {
	                    if ( db === void 0 ) db = databases[ defaultDB ];

	                    return db
	                        .deleteAdmin(adminUsername)
	                        .catch(function (error) {
	                            return new Promise(function (resolve) {
	                                resolve(error);
	                            });
	                        });
	                },
	                disconnect: function disconnect(db) {
	                    if ( db === void 0 ) db = databases[defaultDB];

	                    return new Promise(function (resolve) {
	                        defaultUsername = null;
	                        defaultPassword = null;

	                        if (!db._remote) {
	                            resolve({
	                                message: 'database is not remote',
	                                error: 'bad request',
	                                status: 400,
	                            });
	                            return;
	                        }

	                        db
	                            .logout()
	                            .then(function (res) {
	                                resolve({
	                                    ok: res.ok,
	                                    user: null,
	                                    hasAccess: false,
	                                });
	                            })
	                            .catch(function (error) {
	                                resolve(error);
	                            });
	                    });
	                },

	                destroy: function destroy(db) {
	                    if ( db === void 0 ) db = databases[defaultDB];

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].destroy().then(function () {
	                        if (db !== defaultDB) {
	                            delete databases[db];
	                        }
	                    });
	                },

	                defaults: function defaults(options) {
	                    if ( options === void 0 ) options = {};

	                    pouch.defaults(options);
	                },

	                close: function close(db) {
	                    if ( db === void 0 ) db = databases[defaultDB];

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].close().then(function () {
	                        if (db !== defaultDB) {
	                            delete databases[db];
	                        }
	                    });
	                },

	                getSession: function getSession(db) {
	                    if ( db === void 0 ) db = databases[defaultDB];

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    if (!db._remote) {
	                        return new Promise(function (resolve) {
	                            resolve({
	                                message: 'database is not remote',
	                                error: 'bad request',
	                                status: 400,
	                            });
	                        });
	                    }
	                    return fetchSession();
	                },

	                sync: function sync(localDB, remoteDB, options) {
	                    if (!databases[localDB]) {
	                        makeInstance(localDB);
	                    }
	                    if (!databases[remoteDB]) {
	                        makeInstance(remoteDB, optionsDB);
	                    }
	                    if (!defaultDB) {
	                        defaultDB = databases[remoteDB];
	                    }

	                    var _options = Object.assign(
	                            {},
	                            {
	                                live: true,
	                                retry: true,
	                                back_off_function: function (delay) {
	                                    if (delay === 0) {
	                                        return 1000;
	                                    }
	                                    return delay * 3;
	                                },
	                            },
	                            options
	                        ),
	                        numPaused = 0;

	                    var sync = pouch
	                        .sync(databases[localDB], databases[remoteDB], _options)
	                        .on('paused', function (err) {
	                            if (err) {
	                                vm.$emit('pouchdb-sync-error', {
	                                    db: localDB,
	                                    error: err,
	                                });
	                                return;
	                            }
	                            numPaused += 1;
	                            if (numPaused >= 2) {
	                                vm.$emit('pouchdb-sync-paused', {
	                                    db: localDB,
	                                    paused: true,
	                                });
	                            }
	                        })
	                        .on('change', function (info) {
	                            vm.$emit('pouchdb-sync-change', {
	                                db: localDB,
	                                info: info,
	                            });
	                        })
	                        .on('active', function () {
	                            vm.$emit('pouchdb-sync-active', {
	                                db: localDB,
	                                active: true,
	                            });
	                        })
	                        .on('denied', function (err) {
	                            vm.$emit('pouchdb-sync-denied', {
	                                db: localDB,
	                                error: err,
	                            });
	                        })
	                        .on('complete', function (info) {
	                            vm.$emit('pouchdb-sync-complete', {
	                                db: localDB,
	                                info: info,
	                            });
	                        })
	                        .on('error', function (err) {
	                            vm.$emit('pouchdb-sync-error', {
	                                db: localDB,
	                                error: err,
	                            });
	                        });

	                    fetchSession(databases[remoteDB]);

	                    return sync;
	                },
	                push: function push(localDB, remoteDB, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[localDB]) {
	                        makeInstance(localDB);
	                    }
	                    if (!databases[remoteDB]) {
	                        makeInstance(remoteDB);
	                    }

	                    var numPaused = 0;

	                    var rep = databases[localDB].replicate
	                        .to(databases[remoteDB], options)
	                        .on('paused', function (err) {
	                            if (err) {
	                                vm.$emit('pouchdb-push-error', {
	                                    db: localDB,
	                                    error: err,
	                                });
	                                return;
	                            }
	                            numPaused += 1;
	                            if (numPaused >= 2) {
	                                vm.$emit('pouchdb-push-paused', {
	                                    db: localDB,
	                                    paused: true,
	                                });
	                            }
	                        })
	                        .on('change', function (info) {
	                            vm.$emit('pouchdb-push-change', {
	                                db: localDB,
	                                info: info,
	                            });
	                        })
	                        .on('active', function () {
	                            vm.$emit('pouchdb-push-active', {
	                                db: localDB,
	                                active: true,
	                            });
	                        })
	                        .on('denied', function (err) {
	                            vm.$emit('pouchdb-push-denied', {
	                                db: localDB,
	                                error: err,
	                            });
	                        })
	                        .on('complete', function (info) {
	                            vm.$emit('pouchdb-push-complete', {
	                                db: localDB,
	                                info: info,
	                            });
	                        })
	                        .on('error', function (err) {
	                            vm.$emit('pouchdb-push-error', {
	                                db: localDB,
	                                error: err,
	                            });
	                        });

	                    fetchSession(databases[remoteDB]);

	                    return rep;
	                },

	                pull: function pull(localDB, remoteDB, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[localDB]) {
	                        makeInstance(localDB);
	                    }
	                    if (!databases[remoteDB]) {
	                        makeInstance(remoteDB);
	                    }

	                    var numPaused = 0;

	                    var rep = databases[localDB].replicate
	                        .from(databases[remoteDB], options)
	                        .on('paused', function (err) {
	                            if (err) {
	                                vm.$emit('pouchdb-pull-error', {
	                                    db: localDB,
	                                    error: err,
	                                });
	                                return;
	                            }
	                            numPaused += 1;
	                            if (numPaused >= 2) {
	                                vm.$emit('pouchdb-pull-paused', {
	                                    db: localDB,
	                                    paused: true,
	                                });
	                            }
	                        })
	                        .on('change', function (info) {
	                            vm.$emit('pouchdb-pull-change', {
	                                db: localDB,
	                                info: info,
	                            });
	                        })
	                        .on('active', function () {
	                            vm.$emit('pouchdb-pull-active', {
	                                db: localDB,
	                                active: true,
	                            });
	                        })
	                        .on('denied', function (err) {
	                            vm.$emit('pouchdb-pull-denied', {
	                                db: localDB,
	                                error: err,
	                            });
	                        })
	                        .on('complete', function (info) {
	                            vm.$emit('pouchdb-pull-complete', {
	                                db: localDB,
	                                info: info,
	                            });
	                        })
	                        .on('error', function (err) {
	                            vm.$emit('pouchdb-pull-error', {
	                                db: localDB,
	                                error: err,
	                            });
	                        });

	                    fetchSession(databases[remoteDB]);

	                    return rep;
	                },

	                changes: function changes(db, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    var _options = Object.assign(
	                            {},
	                            {
	                                live: true,
	                                retry: true,
	                                back_off_function: function (delay) {
	                                    if (delay === 0) {
	                                        return 1000;
	                                    }
	                                    return delay * 3;
	                                },
	                            },
	                            options
	                        ),
	                        numPaused = 0;

	                    var changes = db
	                        .changes(_options)
	                        .on('paused', function (err) {
	                            if (err) {
	                                vm.$emit('pouchdb-changes-error', {
	                                    db: localDB,
	                                    error: err,
	                                });
	                                return;
	                            }
	                            numPaused += 1;
	                            if (numPaused >= 2) {
	                                vm.$emit('pouchdb-changes-paused', {
	                                    db: localDB,
	                                    paused: true,
	                                });
	                            }
	                        })
	                        .on('change', function (info) {
	                            vm.$emit('pouchdb-changes-change', {
	                                db: localDB,
	                                info: info,
	                            });
	                        })
	                        .on('active', function () {
	                            vm.$emit('pouchdb-changes-active', {
	                                db: localDB,
	                                active: true,
	                            });
	                        })
	                        .on('denied', function (err) {
	                            vm.$emit('pouchdb-changes-denied', {
	                                db: localDB,
	                                error: err,
	                            });
	                        })
	                        .on('complete', function (info) {
	                            vm.$emit('pouchdb-changes-complete', {
	                                db: localDB,
	                                info: info,
	                            });
	                        })
	                        .on('error', function (err) {
	                            vm.$emit('pouchdb-changes-error', {
	                                db: localDB,
	                                error: err,
	                            });
	                        });

	                    fetchSession(databases[db]);

	                    return changes;
	                },

	                get: function get(db, object, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }
	                    return databases[db].get(object, options);
	                },

	                put: function put(db, object, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }
	                    return databases[db].put(object, options);
	                },

	                post: function post(db, object, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }
	                    return databases[db].post(object, options);
	                },

	                remove: function remove(db, object, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }
	                    return databases[db].remove(object, options);
	                },

	                query: function query(db, fun, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }
	                    return databases[db].query(fun, options);
	                },

	                find: function find(db, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].find(options);
	                },

	                createIndex: function createIndex(db, index) {
	                    if ( index === void 0 ) index = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].createIndex(index);
	                },

	                allDocs: function allDocs(db, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    var _options = Object.assign(
	                        {},
	                        { include_docs: true },
	                        options
	                    );

	                    return databases[db].allDocs(_options);
	                },

	                bulkDocs: function bulkDocs(db, docs, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].bulkDocs(docs, options);
	                },

	                compact: function compact(db, options) {
	                    if ( options === void 0 ) options = {};

	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].compact(options);
	                },

	                viewCleanup: function viewCleanup(db) {
	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].viewCleanup();
	                },

	                info: function info(db) {
	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].info();
	                },

	                putAttachment: function putAttachment(db, docId, rev, attachment) {
	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].putAttachment(
	                        docId,
	                        attachment.id,
	                        rev ? rev : null,
	                        attachment.data,
	                        attachment.type
	                    );
	                },

	                getAttachment: function getAttachment(db, docId, attachmentId) {
	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].getAttachment(docId, attachmentId);
	                },

	                deleteAttachment: function deleteAttachment(db, docId, attachmentId, docRev) {
	                    if (!databases[db]) {
	                        makeInstance(db);
	                    }

	                    return databases[db].removeAttachment(
	                        docId,
	                        attachmentId,
	                        docRev
	                    );
	                },
	            };

	            defineReactive(vm, '$pouch', $pouch);

	            vm.$databases = databases; // Add non-reactive property

	            var pouchOptions = this.$options.pouch;

	            if (!pouchOptions) {
	                return;
	            }

	            if (typeof pouchOptions === 'function') {
	                pouchOptions = pouchOptions();
	            }

	            Object.keys(pouchOptions).map(function (key) {
	                var pouchFn = pouchOptions[key];
	                if (typeof pouchFn !== 'function') {
	                    pouchFn = function () {
	                        return pouchOptions[key];
	                    };
	                }

	                if (typeof vm.$data[key] === 'undefined') {
	                    vm.$data[key] = null;
	                }

	                defineReactive(vm, key, null);
	                vm.$watch(
	                    pouchFn,
	                    function (config) {
	                        if (!config) {
	                            if (!vm[key]) { vm[key] = []; }
	                            return;
	                        }
	                        var selector, sort, skip, limit, first;
	                        if (config.selector) {
	                            selector = config.selector;
	                            sort = config.sort;
	                            skip = config.skip;
	                            limit = config.limit;
	                            first = config.first;
	                        } else {
	                            selector = config;
	                        }

	                        var databaseParam = config.database || key;
	                        var db = null;
	                        if (typeof databaseParam === 'object') {
	                            db = databaseParam;
	                        } else if (typeof databaseParam === 'string') {
	                            if (!databases[databaseParam]) {
	                                databases[databaseParam] = new pouch(
	                                    databaseParam
	                                );
	                                login(databases[databaseParam]);
	                            }
	                            db = databases[databaseParam];
	                        }
	                        if (!db) {
	                            return;
	                        }
	                        if (vm._liveFeeds[key]) {
	                            vm._liveFeeds[key].cancel();
	                        }
	                        var aggregateCache = [];

	                        // the LiveFind plugin returns a liveFeed object
	                        vm._liveFeeds[key] = db
	                            .liveFind({
	                                selector: selector,
	                                sort: sort,
	                                skip: skip,
	                                limit: limit,
	                                aggregate: true,
	                            })
	                            .on('update', function (update, aggregate) {
	                                if (first && aggregate)
	                                    { aggregate = aggregate[0]; }
	                                vm[key] = aggregateCache = aggregate;
	                            })
	                            .on('ready', function () {
	                                vm[key] = aggregateCache;
	                            });
	                    },
	                    {
	                        immediate: true,
	                    }
	                );
	            });
	        },
	    };

	    function installSelectorReplicationPlugin() {
	        // This plugin enables selector-based replication
	        pouch.plugin(function (pouch) {
	            var oldReplicate = pouch.replicate;
	            pouch.replicate = function (source, target, repOptions) {
	                var sourceAjax = source._ajax;
	                source._ajax = function (ajaxOps, callback) {
	                    if (ajaxOps.url.includes('_selector')) {
	                        ajaxOps.url = ajaxOps.url.replace(
	                            'filter=_selector%2F_selector',
	                            'filter=_selector'
	                        );
	                        ajaxOps.method = 'POST';
	                        ajaxOps.body = {
	                            selector: repOptions.selector,
	                        };
	                    }
	                    return sourceAjax(ajaxOps, callback);
	                };
	                return oldReplicate(source, target, repOptions);
	            };
	        });
	    }

	    var api = {
	        mixin: vuePouch,
	        install: function (Vue, options) {
	            vue = Vue;
	            pouch = (options && options.pouch) || PouchDB;
	            installSelectorReplicationPlugin();
	            defaultDB = options && options.defaultDB;

	            if (options.debug) {
	                pouch.debug.enable(options.debug);
	            }

	            // include options for creating databases: https://pouchdb.com/api.html#create_database
	            if (options.optionsDB) {
	                optionsDB = options && options.optionsDB;
	            }

	            Vue.options = Vue.util.mergeOptions(Vue.options, vuePouch);
	        },
	    };

	    module.exports = api;
	})();
	});

	return src;

}));
